//MadExp  5V to 900V converter firmware v0.1
//Copyright 2018(C) Papadopol Lucian Ioan
//------------------------------------------------------------------------------
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//------------------------------------------------------------------------------


#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <SoftwareSerial.h>



#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);

//Define some pins
const int analogInPin = A7;  // Analog input pin
const int interruptPin = 2; // Ingresso impulsi geiger

//Global variables
int cpm = 0;
int Output;
int pulse_count = 0;
unsigned int pastTime = 0;
unsigned int currentTime = 0;
unsigned int timePassed = 0;
int seconds = 60;

//Voltage conversion constant
const float conversionFactor = 0.00488;

//Software serial
SoftwareSerial mySerial(13, 11); //RX jtag 3,TX jtag 5


//Startup setup code
void setup() {
  setupPWM16();

  //Interrupt setup. Active on falling edge of geiger pulse
  pinMode(interruptPin, INPUT);
  attachInterrupt(digitalPinToInterrupt(interruptPin), countPulse, FALLING);

  //Jumper pullups
  pinMode(A0, INPUT_PULLUP);
  pinMode(A1, INPUT_PULLUP);
  pinMode(A2, INPUT_PULLUP);
  pinMode(A3, INPUT_PULLUP);

  //ACK
  pinMode(12, OUTPUT);

  //Start the seria @baudrate specified
  mySerial.begin(9600);


  //OLED init
  // initialize with the I2C addr 0x3C (for the 128x64)
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.display();

  //Take a look at my logo!
  delay(2000);
  display.clearDisplay();
  display.setTextSize(2);
  display.setTextColor(WHITE);

  //Print my info un the UART
  mySerial.println("5V to 900V.max DC-DC converter");
  mySerial.println("by www.madexp.com");
  mySerial.println("(C)Copyright Papadopol Lucian 2018");
  mySerial.println("All rights reserved");
  mySerial.println("");
}


//Main LOOP
void loop() {

  //Reading voltage set point and reading actual converter output voltage
  int Setpoint = 900; //read_jumper();
  int sensorValue = analogRead(analogInPin);
  int Input = sensorValue * conversionFactor * 400;

  //some graphics
  display.setCursor(0, 0);
  display.println(Input);
  display.setCursor(48, 0);
  display.print("V");
  display.setCursor(0, 20);
  display.print(seconds);
  display.setCursor(24, 20);
  display.print("s");
  display.setCursor(48, 20);
  display.print(cpm);
  //display.setCursor(82, 20);
  //display.print("CPM");
  display.setCursor(0, 38);
  display.print(pulse_count);
  display.setCursor(112, 38);
  display.print("P");

  //Write ACK when on tollerance zone

  int setpoint_minus_delta = Setpoint - 20;
  int setpoint_plus_delta = Setpoint + 20;
  if (Input <= setpoint_plus_delta && Input >= setpoint_minus_delta) {
    display.setCursor(92, 0);
    display.print("ACK");
    digitalWrite(12, HIGH);
  }
  if (Input > setpoint_plus_delta || Input < setpoint_minus_delta) {
    display.setCursor(76, 0);
    display.print("NACK");
    digitalWrite(12, LOW);
  }
  display.display();

  //Some calculation for PWM correction (more voltage? less PWM exc...)
  int k = 0 ; //Costante di correzione

  if (Input < Setpoint) {
    k = (Setpoint - Input) / Input;
    Output = Output + 2 + k;
  }
  if (Input > Setpoint) {
    k = (Input - Setpoint) / 10;
    Output = Output - 2 - k;
  }
  analogWrite16(9, Output);
  display.clearDisplay();

  //TIMER 60 sec
  currentTime = millis();
  timePassed = currentTime - pastTime;
  if (timePassed >= 1000) {
    seconds --;
    pastTime = currentTime;
    mySerial.print("#");
  }

  if (seconds == 0) {
    seconds = 60;
    cpm = pulse_count;
    pulse_count = 0;
  }

}

//PWM @ 16bit resolution! From 0 a 4095 not 0-255 like stupid arduino library!!!
/* Configure digital pins 9 and 10 as 16-bit PWM outputs. */
void setupPWM16() {
  DDRB |= _BV(PB1) | _BV(PB2);        /* set pins as outputs */
  TCCR1A = _BV(COM1A1) | _BV(COM1B1)  /* non-inverting PWM */
           | _BV(WGM11);                   /* mode 14: fast PWM, TOP=ICR1 */
  TCCR1B = _BV(WGM13) | _BV(WGM12)
           | _BV(CS10);                    /* no prescaling */
  ICR1 = 0x0fff;                      /* TOP counter value 4095 12bit */
}
/* Comments about the setup
  Changing ICR1 will effect the amount of bits of resolution.
  ICR1 = 0xffff; (65535) 16-bit resolution
  ICR1 = 0x7FFF; (32767) 15-bit resolution
  ICR1 = 0x3FFF; (16383) 14-bit resolution etc....

  Changing the prescaler will effect the frequency of the PWM signal.
  Frequency[Hz}=CPU/(ICR1+1) where in this case CPU=16 MHz
  16-bit PWM will be>>> 16000000/(65535+1)=244,14Hz
  a 12 bit fa 16Mhz/(4095+1)= 3.9Khz
*/
/* 16-bit version of analogWrite(). Works only on pins 9 and 10. */

void analogWrite16(uint8_t pin, uint16_t val)
{
  switch (pin) {
    case  9: OCR1A = val; break;
    case 10: OCR1B = val; break;
  }
}

//Read jumper function
int read_jumper() {
  int point;
  int jumper_presel = 0;
  int jumper_state1 = digitalRead(A0);
  int jumper_state2 = digitalRead(A1);
  int jumper_state3 = digitalRead(A2);
  int jumper_state4 = digitalRead(A3);

  if (jumper_state1 == 0) {
    jumper_presel = jumper_presel + 8;
  }
  if (jumper_state2 == 0) {
    jumper_presel = jumper_presel + 4;
  }
  if (jumper_state3 == 0) {
    jumper_presel = jumper_presel + 2;
  }
  if (jumper_state4 == 0) {
    jumper_presel = jumper_presel + 1;
  }

  point = (jumper_presel * 50) + 200;
  return point;
}

//Pulse counting interrupt. MUST BE SMALLER AS POSSIBLE TO ARCHIVE SPEED!!!
void countPulse()
{
  detachInterrupt(interruptPin); //This is needed!
  pulse_count++;
  attachInterrupt(digitalPinToInterrupt(interruptPin), countPulse, FALLING); //Reactivate interrupt

}
